/***************
Program: POVSUDS.C          (Use medium memory model)
Usage: POVSUDS output.dat
       The screen will turn blue and the speaker will beep when the
       program has completed its full compliment of spheres.

        Basic algorithm from "Computers, Pattern, Chaos, and Beauty" by
        Clifford Pickover (St. Martin's Press).
        Program by Sam Hobbs, C User's Journal, Sept '91

Purpose:
        Draws "osculating" tangent circles of random sizes in two dimensions.
        Outputs raytracer data in the form of user definable objects.

Modification History:
     01/02/92 Tim Wegner
         o Changed to support new PoV {} syntax
         o Added registerVGA call to link with egavga.obj
         o Made errors go to stderr
         o Added 3D support
         o Cleaned up input screen
         o Added multiple textures
         o Added spherical and rectangular bounding options
         o Output to command line parameter
         o Home the screen after <A>gain
     
     11/16/91 Dan Farmer
         o Added PoV raytrace output.
         o (Uses printf() with redirection to create output include file.)
         o Track object extents and auto-center object.
         o Get user input for variable parameters.
         o Add z to circle_type structure.  (Not used yet.)
         o Add PoV Noise and Turbulence variations

Wishlist/Helplist:

 If you have the desire and knowledge to enhance this program, please feel
 free to do so, but also, please share your efforts with us.  Here are some
 of the things that I would like to see added:
-   Better interface.

-   Portability.
        - graphics calls... move to  #defines:
            - circle()
            - init_graphics()
            - closegraph()
            - rectangle()
            - cleardevice()
        - use PoV graphics calls.
        - any others?

-   Write Vivid 2.0 data files.

-   3D versions

-   Contain all objects within a bounding volume
         (Can this program be used to fill a jar with jawbreakers?)

-   Add effects of gravity or spin, etc.

-   Use various functions for determining object's height position.
        - use rand, turb or noise for height?

    (see letter from Drew, below)

Hi Dan,
  I really liked your suds image. I wanted to write a similar program a couple
of months ago, but times are busy :).

  I had a couple of ideas that you might want to incorporate into your suds
program. If you use quadric spheres you could scale the Y (up) component to be
the same on all the ellipsoids. Add a textured X-Z plane through the center of
these and you'd get a cobblestone effect.

  You could also make a "suds mountain" by scaling the Y component inversely to
the sphere's distance to the center of the first sphere. That is, the each
sphere's Y component would decrease as they got farther away from the center.
You could get into fractals too, but the suds cone is a nice easy trick.

       --Drew

****************/
#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <math.h>
#include <time.h>       /* required for random function */
#include <dos.h>


#define FALSE 0
#define TRUE !FALSE

#define RANDOMIZE randomize()
#define RAND rand()


#define DBL double

#define MIN_RADIUS 0.5
#define MIN_CIRCLES 1
#define MAX_CIRCLES 2048         /* This is really quite arbitrary. dmf */
#define MAX_RADIUS_DIVISOR 5 /* can be changed to allow different max radii*/

DBL distance( DBL x, DBL y, DBL z, int i);
int main(int argc, char *argv[]);
void init_graphics (void);
void generate_circles();
void get_user_input();
void exit_if_kbhit();
void write_pov_data(void);
void write_pov_object(int x,int y,int z,int r);
void movetoxy(int x, int y);

typedef struct {
    int x;
    int y;
    int z;
    int r;
    } circle_type;

circle_type c[MAX_CIRCLES+1];
int num_circles = 200;
float min_radius = .5;
int max_radius_divisor =  5;
int N, rmax;
int textures=0;
int raytrace = 1;
int xmax, ymax; /* currently set to screen coordinates in get_user_input() */
int zmax = 0;
int constrain = 0;
FILE *fpout;

int main (int argc, char *argv[])
{
    int keypress,done;
    if(argc != 2) {
       printf("\nPOV Suds\n  Usage: POVSUDS <outfile>\n");
       exit(1);
    }
    if((fpout=fopen(argv[1],"w"))==NULL) {
       printf("Can't open %s ... aborting\n",argv[1]);
       exit(1);
    }   
    
    done=FALSE;

    init_graphics();
    do{
        setbkcolor(0);                                  /* color to black */
        cleardevice();
        get_user_input();

        cleardevice();
        generate_circles();

        /* Frame the screen to show user that program is finished */
        fputs("\007",stdout);                                 /* beep */
        fputs("<W>rite <Q>uit <A>gain",stdout);
        rectangle(0,0,getmaxx(),getmaxy());             /* draw frame */
        setbkcolor(1);                                  /* color to blue */

        keypress = getch();                             /* wait for keypress */
        switch (keypress)
        {
        case 'A':
        case 'a':
            done = FALSE;
            break;
        case 'W':
        case 'w':
            write_pov_data();                           /* write PoV data */
            closegraph();
            return(0);                           /* quit */
        case 'Q':
        case 'q':
            done = TRUE;
            break;
        default:
            break;
        }
    } while (!done);
    
    closegraph();
    return 0;
}


void generate_circles()
{
    int  x_pos, y_pos, z_pos, radius;
    RANDOMIZE;

    /* Note that xmax and ymax would probably be better for raytrace
       output if defined as a radius instead of screen rez.  This would
       clump the circles into a larger circular area.    */

    rmax = max(xmax,ymax) / max_radius_divisor;

    /* constrain to this sphere - tucked safely at the end of c[] */
    c[MAX_CIRCLES].x = xmax/2;
    c[MAX_CIRCLES].y = ymax/2;
    c[MAX_CIRCLES].z = zmax/2;
    c[MAX_CIRCLES].r = min(xmax,ymax) * 0.5;

    /* draw the rest of the circles */
    for( N = 0; N <= num_circles; N++) {
        do {
            exit_if_kbhit();
            x_pos = RAND % xmax;
            y_pos = RAND % ymax;
            if(zmax > 0)
                z_pos = RAND % zmax;
            else
                z_pos = 0;
                radius = new_radius(x_pos, y_pos, z_pos);
        } while (radius <= min_radius);
        radius = min( rmax, radius);
        circle(x_pos,y_pos, radius);
        c[N].x = x_pos;
        c[N].y = y_pos;
        c[N].z = z_pos;
        c[N].r = radius;
    }
}

/* Function: new_radius(x,y,z)
   Purpose:
   Returns the distance from the point x,y,z to the nearest circle.
   Aborts and returns the calculated negative or zero distance if x,y
   lies on or inside of another circle
*/

#define OLD     1
#define BOX     2
#define SPHERE  3

int new_radius(int x, int y, int z)
{
    int i, radius,dist_circle;

    radius = rmax;
    switch(constrain)
    {
    case OLD:
       break;
    case BOX:
        /* don't let get outside box */
        if(radius > x)
            radius = x;
        if(radius > xmax - x)
           radius = xmax - x;
        if(radius > y)
           radius = y;
        if(radius > ymax - y)
           radius = ymax - y;
        if(zmax > min_radius*2) {
            if(radius > z)
                radius = z;
            if(radius > zmax - z)
                radius = zmax - z;
        }      
        break;
    case SPHERE:
       dist_circle = distance(x,y,z,MAX_CIRCLES); /* dist pt to center */
       if(dist_circle >= 0) 
          return radius = 0; /* center ouside bounding sphere */
       else
          radius = min(radius,-dist_circle);
       break;
    }
    if(N>0) /* don't do this the first time */
        for (i = 0; i <= N; i++) {
            dist_circle = (int) distance(x,y,z,i);
            radius = min(radius, dist_circle);
            if (radius <= min_radius ) return radius;
        }
    return radius;
}


/*  Function: distance(x,y,z,i)
        Purpose:
        Returns the distance from the point x,y,z to edge of circle i.
        Returns a negative value or zero distance if x,y is inside or on
        circle i.
*/
DBL distance(DBL x, DBL y, DBL z, int i)
{
    DBL xdist, ydist, zdist, distancesq,dist;
    xdist = x - c[i].x;
    ydist = y - c[i].y;
    zdist = z - c[i].z;
    distancesq = (xdist*xdist) + (ydist*ydist) + (zdist*zdist);
    dist=sqrt(distancesq)-c[i].r;
    return (dist);
}

/*
   Get parameters from user.
*/
static char *tracers[] = {"PoV Version 1.00","PoV Version .5"};
#define NUM_TRACERS (sizeof(tracers)/sizeof(char *))

static char *bounds[] = {"centers constrained to rectangular region",
                         "objects constrained to rectangular region",
                         "objects constrained to spherical region"};
#define NUM_BOUNDS (sizeof(bounds)/sizeof(char *))

void get_user_input()
{
    /* Get user input */
    /* Note that the ranges here are pretty much arbitrary values at this
       time. Need to test to establish good defaults and ranges. dmf */
    int i;
    char buf[200];

    /* center center boundaries to screen coordinates */
    xmax = getmaxx();
    ymax = getmaxy();
    movetoxy(0,0);
    printf("\nPOV SUDS - Generate Osculating Objects\n\n");
    printf("Press <Return> to keep same values\n");
    printf("Current values shown in []\n");

    printf("\nEnter Ray Tracer\n");
    for(i=0;i<NUM_TRACERS;i++)
       printf("%-d=%s\n",i+1,tracers[i]);    
    printf("Choice [%-d] : ",raytrace);
    gets(buf);
    sscanf(buf,"%d->",&raytrace);
    if(raytrace < 1 || raytrace > NUM_TRACERS )
       raytrace=1;

    printf("\nEnter Type of Bounds\n");
    for(i=0;i<NUM_BOUNDS;i++)
       printf("%-d=%s\n",i+1,bounds[i]);    
    printf("Choice [%-d] : ",constrain);
    gets(buf);
    sscanf(buf,"%d->",&constrain);
    if(constrain < 1 || constrain > NUM_BOUNDS )
       constrain=1;

    printf("\nRegion Containing Spheres:\n");
    printf("   X from 0 to %-3d (fixed to screen resolution)\n",xmax);
    printf("   Y from 0 to %-3d (fixed to screen resolution)\n",ymax);
    printf("   Z from 0 to zmax (user set)\n");
    printf("Enter zmax (depth of region ) [%-d] : ",zmax);
    gets(buf);
    sscanf(buf,"%d->",&zmax);
    if(zmax < 0) zmax = 0;

    printf("\nEnter Number Textures [%-d] : ",textures);
    gets(buf);
    sscanf(buf,"%d->",&textures);
    if(textures < 0)
       textures = 0;

    printf("\nEnter Max Spheres (1-2048) [%-d] : ",num_circles);
    gets(buf);
    sscanf(buf,"%d",&num_circles);
    if (num_circles < MIN_CIRCLES)  num_circles = MIN_CIRCLES;
    if (num_circles > MAX_CIRCLES)  num_circles = MAX_CIRCLES;

    printf("\nEnter Min Radius (0.0001-10) [%-7.4f] : ",min_radius);
    gets(buf);
    sscanf(buf,"%f",&min_radius);
    if (min_radius < 0.0001)   min_radius = 0.0001;
    if (min_radius > 10.0)     min_radius = 10.0;

    printf("\nEnter Max Radius Divisor (1-100) [%-d] : ",max_radius_divisor);
    gets(buf);
    sscanf(buf,"%d",&max_radius_divisor);
    if (max_radius_divisor < 1)   max_radius_divisor = 1;
    if (max_radius_divisor > 100) max_radius_divisor = 100;
}

void exit_if_kbhit()
{
    if (kbhit()) {
        getch();
        closegraph();
        exit(1);
    }
}

void gerror(int errorcode)
{
        fprintf(stderr,"Graphics error: %s\n", grapherrormsg(errorcode));
        fprintf(stderr,"Press any key to halt.");
        getch();
        exit(1);
}

/* Uses the Borland BGI drivers */
void init_graphics (void)
{
    int gdriver = DETECT, gmode, errorcode;
    /* cut following 3 lines if you want to dynamically load egavga.bgi */
    errorcode = registerbgidriver(EGAVGA_driver);
    if (errorcode < 0)
        gerror(errorcode);

    initgraph(&gdriver, &gmode,"");
    errorcode = graphresult();
    if (errorcode != grOk)
        gerror(errorcode);
}


/*
   Write an include file for PoV Ray as a composite of declared objects.
   The composite is on the X/Y plane, left-handed coordinate system.
   Uses a main datafile that defines "SudsObject" and an associated texture.

    Format: 
    (POV 1.00)
    #declare Suds = composite {
        object { SudsObject texture { suds0 } scale <x y z> translate <x y z> }
        object { SudsObject texture { suds1 } scale <x y z> translate <x y z> }
        object { SudsObject texture { suds2 } scale <x y z> translate <x y z> }
    }

    (POV .5)
    #declare Suds = composite 
        object SudsObject texture suds0 end_texture scale <x y z> translate <x y z> end_texture
        object SudsObject texture suds1 end_texture scale <x y z> translate <x y z> end_texture
        object SudsObject texture suds2 end_texture scale <x y z> translate <x y z> end_texture
    }

*/
void write_pov_data(void)
{
        int i;
        int maxr_extent = 0;
    int maxx_extent = 0;
    int maxy_extent = 0;
    int maxz_extent = 0;
    int minx_extent = 1.7e+10;
        int miny_extent = 1.7e+10;
        int minz_extent = 1.7e+10;
    int len_x,len_y,len_z,half_x, half_y,half_z,
       x_tran,y_tran,z_tran,x_len,y_len,z_len;
    int x_pos,y_pos,z_pos,radius;

    switch(raytrace)
    {
       case 1:
          fprintf(fpout,"#declare Suds = composite {\n");
          break;
       case 2:
          fprintf(fpout,"#declare Suds = composite \n");
          break;
    }      
          
    /* Write out each of the objects */
    for(i=0;i < N; i++) {
        x_pos  =   c[i].x;
        y_pos  =   c[i].y;
        z_pos  =   c[i].z;
        radius =   c[i].r;

        write_pov_object(x_pos,y_pos,z_pos,radius);

        maxx_extent = (int) max(x_pos+radius,maxx_extent);
        maxy_extent = (int) max(y_pos+radius,maxy_extent);
        maxz_extent = (int) max(z_pos+radius,maxz_extent);
        minx_extent = (int) min(x_pos-radius,minx_extent);
        miny_extent = (int) min(y_pos-radius,miny_extent);
        minz_extent = (int) min(z_pos-radius,minz_extent);
        maxr_extent = (int) max(radius,maxr_extent);
    }

    /* Center the composite object, based on extents */
    x_len = maxx_extent + abs(minx_extent);
    y_len = maxy_extent + abs(miny_extent);
    z_len = maxz_extent + abs(minz_extent);
    half_x = x_len * 0.5;
    half_y = y_len * 0.5;
    half_z = z_len * 0.5;
        x_tran = half_x + minx_extent;
        y_tran = half_y + miny_extent;
        z_tran = half_z + minz_extent;

    switch(raytrace)
    {
    case 1: /* POV 1.00 */
       fprintf(fpout,"    translate <%5d %5d %5d>\n",-x_tran,-y_tran,-z_tran);
       fprintf(fpout,"}\n");
       fprintf(fpout,"\n");
       /* note: in new PoV Cubes have side 1 */
       fprintf(fpout,"#declare Bounding_Box = intersection { Cube\n");
       fprintf(fpout,"        scale < %d %d %d >\n",x_len,y_len,z_len);
       fprintf(fpout,"    }\n");
       fprintf(fpout,"\n");
    
       fprintf(fpout,"//Object Extents:  min x = %5d min y = %5d min z = %5d \n",minx_extent,miny_extent,minz_extent);
       fprintf(fpout,"//                 max x = %5d max y = %5d max z = %5d \n",maxx_extent,maxy_extent,maxz_extent);
       fprintf(fpout,"//                 max radius = %d \n",maxr_extent);
       fprintf(fpout,"//                 %d objects generated\n",N);
       break;
    case 2: /* POV beta .5 */
       fprintf(fpout,"    translate <%5d %5d %5d>\n",-x_tran,-y_tran,-z_tran);
       fprintf(fpout,"end_composite\n");
       fprintf(fpout,"\n");
       fprintf(fpout,"#declare Bounding_Box = intersection Cube\n");
       fprintf(fpout,"        scale < %d %d %d >\n",half_x,half_y,half_z);
       fprintf(fpout,"    end_intersection\n");
       fprintf(fpout,"\n");
    
       fprintf(fpout,"{Object Extents:  min x = %5d min y = %5d min z = %5d \n",minx_extent,miny_extent,minz_extent);
       fprintf(fpout,"                 max x = %5d max y = %5d max z = %5d \n",maxx_extent,maxy_extent,maxz_extent);
       fprintf(fpout,"                 max radius = %d \n",maxr_extent);
       fprintf(fpout,"                 %d objects generated}\n",N);

       break;
    }           
    fprintf(fpout,"\n// POV SUDS settings used to generate this file\n\n");
    fprintf(fpout,"// Ray Tracer         = %d (%s)\n",raytrace,tracers[raytrace-1]);
    fprintf(fpout,"// Type of Bounds = %d (%s)\n",constrain,bounds[constrain-1]);
    fprintf(fpout,"// Region Containing Objects:\n");
    fprintf(fpout,"//    X from 0 to %-3d (fixed to screen resolution)\n",xmax);
    fprintf(fpout,"//    Y from 0 to %-3d (fixed to screen resolution)\n",ymax);
    fprintf(fpout,"//    Z from 0 to %-3d (user setting of zmax)\n",zmax);
    fprintf(fpout,"// Number of Textures = %d\n",textures);
    fprintf(fpout,"// Max Spheres        = %d\n",num_circles);
    fprintf(fpout,"// Min Radius         = %f\n",min_radius);
    fprintf(fpout,"// Max Radius Divisor = %d\n",max_radius_divisor);
}

/* Write a single circle out to PoV Ray format as a user-definable object */
void write_pov_object(int x, int y, int z, int r)
{
    static ct=0;
    int x_scale = r;
    int y_scale = x_scale;
    int z_scale = x_scale;

    int x_trans = x;
    int y_trans = y;
    int z_trans = z;

    switch(raytrace)
    {
    case 1:
       if(textures) {
          fprintf(fpout,"    object { SudsObject TEXTURE {suds%-3d} scale <%3d %3d %3d> translate <%3d %3d %3d> }\n",ct,x_scale,y_scale,z_scale,x_trans,y_trans,z_trans);
       ct++;
       if(ct==textures)
         ct=0;
       } else
          fprintf(fpout,"    object { SudsObject scale <%3d %3d %3d> translate <%3d %3d %3d> }\n",x_scale,y_scale,z_scale,x_trans,y_trans,z_trans);
       break;
    case 2:
       if(textures) {
          fprintf(fpout,"    object SudsObject texture suds%-3d end_texture scale <%3d %3d %3d> translate <%3d %3d %3d> end_object\n",ct,x_scale,y_scale,z_scale,x_trans,y_trans,z_trans);
       ct++;
       if(ct==textures)
         ct=0;
       } else
          fprintf(fpout,"    object SudsObject scale <%3d %3d %3d> translate <%3d %3d %3d> end_object\n",x_scale,y_scale,z_scale,x_trans,y_trans,z_trans);
       break;
   }
}
#define VIDEO 0x10

void movetoxy(int x, int y)
{
   union REGS regs;
   regs.h.ah = 2;
   regs.h.dh = y;
   regs.h.dl = x;
   regs.h.bh = 0;
   int86(VIDEO,&regs,&regs);
}
